#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <string.h>

//Mode0
//BG2 \
//BG3 ͕
//Spr J[\

//---------------------------------------------------------------------------
//
#define BG_CX					8
#define BG_CY					8
#define BG_SCREEN_CX			30
#define BG_SCREEN_CY			20
#define BG_WORLD_SCREEN_CX		32
#define BG_WORLD_SCREEN_CY		32
#define BG_WORLD_SCREEN_WCX		256
#define BG_WORLD_SCREEN_WCY		256
#define BG_WRITE_EY				(20-3)

//is
#define STATE_INIT				1
#define STATE_INPUT				2
#define STATE_WRITE				3
#define STATE_CALC				4

//̓L[̃EFCg
#define KEY_REPEAT_TIME			500
#define TIMER_INTMS				10

#define MAX_STR_LENGTH			(64-1)	//͕̍ő吔
#define MAX_LABEL_LENGTH		(32-1)
#define MAX_CMD_LENGTH			(16-1)

#define MAX_SYMBOL_TABLE_CNT	32
#define MAX_CMD_TABLE_CNT		2

//g[N
#define EOS						257		//̋؂L
#define INT						258		//l
#define SYMBOL					259		//V{
#define EQ						260		//==
#define NE						261		//!=
#define LE						262		//<=
#define GE						263		//>=
#define _IF						264		//ߌ _if
#define _WHILE					265		//ߌ _while

//---------------------------------------------------------------------------
typedef struct {
	u8 Chr;
	u8 Map;
	u8 Col;
} ST_BG;

typedef struct {
	s16 x;
	s16 y;
	u16 cx;
	u16 cy;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_WRITER;

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	s16 minX;		//ړ
	s16 maxX;
	s16 minY;
	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
	u8  keyNow;		//ړp
	u8  keyOld;		//1O̓
} ST_CURSOR;

typedef struct {
	u16 type;
	union {
		s32 val;
		u8* str;
	} u;
} ST_TOKEN;

typedef struct {
	u16 type;
	u8  str[MAX_LABEL_LENGTH];
	s32 val;
} ST_SYMBOL;

typedef struct {
	u16 type;
	u8  str[MAX_CMD_LENGTH];
} ST_CMD;

typedef struct {
	u8  str[MAX_STR_LENGTH];
	s16 readCnt;
	s16 writeCnt;
} ST_BUFFER;

typedef struct {
	u32 clockTick;
	u32 wait;
	u32 repeat;
} ST_TIMER;

//---------------------------------------------------------------------------
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
ST_WRITER Writer;
ST_TOKEN  Token;
ST_BUFFER Buffer;
ST_TIMER  Timer;
ST_SYMBOL SymTbl[MAX_SYMBOL_TABLE_CNT];
ST_CMD    CmdTbl[MAX_CMD_TABLE_CNT];

u8  State;			//istO
u16 BGvofs;			//XN[p
s16 SymRegistCnt;
//---------------------------------------------------------------------------
void StateInit();
void StateInput();
void StateWrite();
void StateCalc();

void InitWS();
void InitBG();
void InitSprite();
void InitCursor();
void InitBuffer();
void InitTimer();
void InitToken();
void IRQUserHandler();

void DrawFrame();
void DrawCursor();
void DrawWriter();
void DrawAnswer(s32 ans);

void ProcKeyInput();
void ProcDispWrite();
void ProcScroll();
void ProcBGLineFeed();
s32  ProcCalc();

bool WaitTimer();
bool RepeatKey();
bool ChkMoveCursor();
void MoveCursor();

void     lexar();
s32      lexarGetNumber();
u8*      lexarGetSymbol();
u16      lexarGetCmdType();
u16      lexarGetSignType(u16 type);
ST_TOKEN program();
ST_TOKEN stmt();
ST_TOKEN stmt_if();
ST_TOKEN stmt_while();
ST_TOKEN expr();
ST_TOKEN expr2();
ST_TOKEN simpleExpr();
ST_TOKEN simpleExpr2();
ST_TOKEN term();
ST_TOKEN term2();
ST_TOKEN factor();
ST_TOKEN Run(u16 op, ST_TOKEN code1, ST_TOKEN code2);
bool     IsMatch(ST_TOKEN t, u16 type);
bool     IsTrue(ST_TOKEN t);
bool     IsNumber(u8 chr);
bool     IsName(u8 chr);
bool     IsCmd(u8 chr);
bool     HasSymbol(u8* str);
u8*      AddSymbol(u8* str);
s32      GetSymbolVal(u8* str);
u8*      GetSymbolStr(u8* str);
u16      GetSymbolNum(u8* str);
u8*      GetSymbolPointer(u8* str);
bool     ChgSymbol(u8* str, s32 val);

//---------------------------------------------------------------------------
int main(void)
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();   break;
			case STATE_INPUT: StateInput();  break;
			case STATE_WRITE: StateWrite();  break;
			case STATE_CALC:  StateCalc();   break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	KeyInit();

	InitWS();
	InitBG();
	InitSprite();
	InitCursor();
	InitBuffer();
	InitTimer();
	InitToken();

	DrawFrame();
	DrawCursor();
	DrawWriter();

	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
//EFCgRg[
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG2 \
	//BG3 ͕

	BG2.Chr = 1;
	BG2.Map = 0;
	BG2.Col = 0;

	BG3.Chr = 1;	//BG2Ɠ
	BG3.Map = 1;
	BG3.Col = 0;	//BG2Ɠ

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.Chr) | BG_MAPBASE(BG2.Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 3);

	BGSetPal16  (BG3.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG3.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();
	SpriteSetPal16((u16*)&cursor_pal_bin);
	SpriteSetData ((u16*)&cursor_img_bin, CURSOR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitCursor()
{
	//\p
	Writer.x       = 0;
	Writer.y       = 0;
	Writer.cx      = 8;
	Writer.cy      = 8;
	Writer.sprNo   = 1;
	Writer.sprTile = 0;
	Writer.sprSize = SP_SIZE_16;
	Writer.sprForm = SP_SQUARE;

	SpriteSetSize(Writer.sprNo, Writer.sprSize, Writer.sprForm, SP_COLOR_16);
	SpriteSetChr (Writer.sprNo, Writer.sprTile);

	//͗p
	Cursor.x       =  0;
	Cursor.y       = 18;
	Cursor.cx      =  8;
	Cursor.cy      =  8;
	Cursor.minX    =  0;			//ړ
	Cursor.maxX    = 25;
	Cursor.minY    = 18;
	Cursor.maxY    = 19;
	Cursor.sprNo   =  2;
	Cursor.sprTile =  0;
	Cursor.sprSize = SP_SIZE_16;
	Cursor.sprForm = SP_SQUARE;

	SpriteSetSize(Cursor.sprNo, Cursor.sprSize, Cursor.sprForm, SP_COLOR_16);
	SpriteSetChr (Cursor.sprNo, Cursor.sprTile);
}
//---------------------------------------------------------------------------
void DrawFrame()
{
	BGDrawCls(BG3.Map, ' ');

	BGDrawStr(BG2.Map, 0, 17, (u8*)"------------------------------");
	BGDrawStr(BG2.Map, 0, 18, (u8*)"0123456789 +-*/%()<> !=_ ;    ");
	BGDrawStr(BG2.Map, 0, 19, (u8*)"abcdefghijklmnopqrstuvwxyz    ");
}
//---------------------------------------------------------------------------
void DrawCursor()
{
	s16 x  = Cursor.x * Cursor.cx;
	s16 y  = Cursor.y * Cursor.cy;
	u16 no = Cursor.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
void DrawWriter()
{
	s16 x  = Writer.x * Writer.cx;
	s16 y  = Writer.y * Writer.cy;
	u16 no = Writer.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
void InitBuffer()
{
	s16 i;
	for(i=0; i<MAX_STR_LENGTH; i++)
	{
		Buffer.str[i] = ' ';
	}
	Buffer.readCnt  = 0;
	Buffer.writeCnt = 0;
}
//---------------------------------------------------------------------------
void InitTimer()
{
	Timer.wait      = 0;
	Timer.repeat    = 0;
	Timer.clockTick = 0;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;

	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void StateInput()
{
	ProcKeyInput();
}
//---------------------------------------------------------------------------
void StateWrite()
{
	ProcDispWrite();
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	//^C}[EFCg
	if( WaitTimer() == TRUE )
	{
		return;
	}

	ST_CURSOR* p = &Cursor;
	p->keyNow    = KeyInput();

	//s[h
	if( RepeatKey() == TRUE )
	{
		return;
	}

	if( ChkMoveCursor() == TRUE )
	{
		MoveCursor();
		DrawCursor();
	}

	if(p->keyNow & KEY_A)
	{
		State = STATE_WRITE;
	}
}
//---------------------------------------------------------------------------
//L[s[h
bool RepeatKey()
{
	ST_CURSOR* p = &Cursor;
	ST_TIMER*  t = &Timer;

	if(p->keyNow && p->keyNow == p->keyOld)
	{
		if(t->clockTick - t->repeat < KEY_REPEAT_TIME)
		{
			return TRUE;
		}

		return FALSE;
	}

	p->keyOld = p->keyNow;
	t->repeat = t->clockTick;

	return FALSE;
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
bool ChkMoveCursor()
{
	ST_CURSOR* p = &Cursor;

	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->keyNow & KEY_UP   ) ty1--;
	if(p->keyNow & KEY_DOWN ) ty1++;
	if(p->keyNow & KEY_LEFT ) tx1--;
	if(p->keyNow & KEY_RIGHT) tx1++;

	if(tx1 < p->minX || ty1 < p->minY) return FALSE;
	if(tx1 > p->maxX || ty1 > p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
//J[\̈ړ
void MoveCursor()
{
	ST_CURSOR* p = &Cursor;

	if(p->keyNow & KEY_UP   ) p->y--;
	if(p->keyNow & KEY_DOWN ) p->y++;
	if(p->keyNow & KEY_LEFT ) p->x--;
	if(p->keyNow & KEY_RIGHT) p->x++;
}
//---------------------------------------------------------------------------
void ProcDispWrite()
{
	ST_WRITER* wp = &Writer;
	ST_CURSOR* cp = &Cursor;
	u16 chr       = BGGetChar(BG2.Map, cp->x, cp->y); 	//J[\WLN^擾

	//擾LN^`
	BGDrawChar(BG3.Map, wp->x, (wp->y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, BG3.Col, chr);
	Buffer.str[Buffer.writeCnt++] = chr;

	//܂Ԃ
	ProcBGLineFeed();

	if(chr == ';')
	{
		Writer.x = BG_SCREEN_CX;	//s܂
		ProcBGLineFeed();

		State = STATE_CALC;
	}
	else
	{
		State = STATE_INPUT;
	}
}
//---------------------------------------------------------------------------
void ProcScroll()
{
	BGvofs = (BGvofs + BG_CY) % BG_WORLD_SCREEN_WCY;

	//Psɏꂽe܂
	BGDrawStr(BG3.Map, 0, (Writer.y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, (u8*)"                              ");

//	REG_BG3HOFS = BGhofs;		//x
	REG_BG3VOFS = BGvofs;		//y
}
//---------------------------------------------------------------------------
void StateCalc()
{
	s32 ans = ProcCalc();
	DrawAnswer(ans);

	InitBuffer();

	Writer.x = BG_SCREEN_CX;	//s܂
	ProcBGLineFeed();

	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
s32 ProcCalc()
{
	ST_TOKEN t;

	lexar();
	t = program();

	return t.u.val;
}
//---------------------------------------------------------------------------
void lexar()
{
	u8 chr;
	for(;;)
	{
		chr = Buffer.str[Buffer.readCnt++];

		if(chr != ' ')
		{
			break;
		}
	}

	switch(chr)
	{
	case '+':
	case '-':
	case '*':
	case '/':
	case '(':
	case ')':
	case ';':
		Token.type = chr;
		return;

	case '=':
		Token.type = lexarGetSignType(EQ); // = or ==
		return;

	case '!':
		Token.type = lexarGetSignType(NE); // ! or !=
		return;

	case '<':
		Token.type = lexarGetSignType(LE); // < or <=
		return;

	case '>':
		Token.type = lexarGetSignType(GE); // > or >=
		return;

	default:
		if( IsCmd(chr) == TRUE ) //ߌ`FbN
		{
			Token.type = lexarGetCmdType();
			return;
		}
		else if( IsName(chr) == TRUE ) //p`FbN
		{
			Token.type  = SYMBOL;
			Token.u.str = lexarGetSymbol();
			return;
		}
		else if( IsNumber(chr) == TRUE ) //l`FbN
		{
			Token.type  = INT;
			Token.u.val = lexarGetNumber();
			return;
		}
		else
		{
			Token.type = EOS;
			return;
		}
	}
}
//---------------------------------------------------------------------------
s32 lexarGetNumber()
{
	u8  chr = Buffer.str[Buffer.readCnt-1];
	s32 val = 0;

	do
	{
		val = (val * 10) + (chr - '0');		//ӂ`FbNĂȂł
		chr = Buffer.str[Buffer.readCnt++];

	} while( IsNumber(chr) == TRUE );
	Buffer.readCnt--;

	return val;
}
//---------------------------------------------------------------------------
u8* lexarGetSymbol()
{
	u8  chr = Buffer.str[Buffer.readCnt-1];
	u8  str[MAX_LABEL_LENGTH];
	s16 i;
	
	i = 0;
	do
	{
		str[i++] = chr;		//ӂ`FbNĂȂł
		chr      = Buffer.str[Buffer.readCnt++];

	} while( IsName(chr) == TRUE );
	str[i] = '\0';
	Buffer.readCnt--;

	//ɃV{͓o^ς݂`FbN܂
	if( HasSymbol(str) == TRUE )
	{
		return GetSymbolPointer(str);
	}

	return AddSymbol(str);
}
//---------------------------------------------------------------------------
u16 lexarGetCmdType()
{
	u8  chr = Buffer.str[Buffer.readCnt-1];
	u8  str[MAX_LABEL_LENGTH];
	s16 i;
	
	i = 0;
	do
	{
		str[i++] = chr;		//ӂ`FbNĂȂł
		chr      = Buffer.str[Buffer.readCnt++];

	} while( IsName(chr) == TRUE || IsCmd(chr) == TRUE );
	str[i] = '\0';
	Buffer.readCnt--;

	//o^Ă閽ߌƔr܂
	for(i=0; i<MAX_CMD_TABLE_CNT; i++)
	{
		if( strcmp((char*)str, (char*)CmdTbl[i].str) == 0 )
		{
			return CmdTbl[i].type;
		}
	}

	return EOS;
}
//---------------------------------------------------------------------------
u16 lexarGetSignType(u16 type)
{
	u8  bakChr = Buffer.str[Buffer.readCnt-1];
	u8  chr    = Buffer.str[Buffer.readCnt++];

	if(chr == '=')
	{
		return type;
	}

	Buffer.readCnt--;
	return bakChr;
}
//---------------------------------------------------------------------------
bool IsName(u8 chr)
{
	return (chr >= 'a' && chr <= 'z') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool IsNumber(u8 chr)
{
	return (chr >= '0' && chr <= '9') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool IsCmd(u8 chr)
{
	return (chr == '_') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool IsTrue(ST_TOKEN t)
{
	return (t.u.val != 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool IsMatch(ST_TOKEN t, u16 type)
{
	return t.type == type ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
ST_TOKEN program()
{
	ST_TOKEN t;

	t = stmt();

	switch(Token.type)
	{
	case ';':
		break;
	default:
		t.type  = 0;
		t.u.val = 0;
		break;
	}

	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN stmt()
{
	switch(Token.type)
	{
	case _IF:
		return stmt_if();

	case _WHILE:
		return stmt_while();

	default:
		return expr();
	}
}
//---------------------------------------------------------------------------
ST_TOKEN stmt_if()
{
	ST_TOKEN miss;
	miss.type = EOS;

	lexar();
	if(Token.type != '(')
	{
		return miss;
	}

	lexar();
    ST_TOKEN t = expr();

	if(Token.type != ')')
	{
		return miss;
	}

	if( IsTrue(t) == FALSE )
	{
		return miss;
	}

	lexar();
	return stmt();
}
//---------------------------------------------------------------------------
ST_TOKEN stmt_while()
{
	ST_TOKEN miss;
	miss.type  = EOS;
	miss.u.val = 0;

	ST_TOKEN ret = miss;
	u16 readCnt  = Buffer.readCnt;	//[v̍łɎgp܂

st_while:

	lexar();
	if(Token.type != '(')
	{
		return miss;
	}

	lexar();
    ST_TOKEN t = expr();

	if(Token.type != ')')
	{
		return miss;
	}

	if( IsTrue(t) == TRUE )
	{
		lexar();
		ret = stmt();

		Buffer.readCnt = readCnt;
		goto st_while;
	}

	Token.type  = ';';
	return ret;
}
//---------------------------------------------------------------------------
ST_TOKEN expr()
{
	ST_TOKEN t = simpleExpr();

	switch(Token.type)
	{
	case '<':
	case '>':
	case EQ:
	case NE:
	case LE:
	case GE:
		return expr2(t);
	default:
		return t;
	}
}
//---------------------------------------------------------------------------
ST_TOKEN expr2(ST_TOKEN t)
{
	while(Token.type == '<' || Token.type == '>' ||
	      Token.type == EQ  || Token.type == NE  ||
	      Token.type == LE  || Token.type == GE)
	{
		u16 type = Token.type;

		lexar();
		ST_TOKEN t2 = simpleExpr();

		if( IsMatch(t2, INT) == TRUE )
		{
			t = Run(type, t, t2);
		}
	}

	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN simpleExpr()
{
	ST_TOKEN t = term();

	switch(Token.type)
	{
	case '+':
	case '-':
		return simpleExpr2(t);
	default:
		return t;
	}

}
//---------------------------------------------------------------------------
ST_TOKEN simpleExpr2(ST_TOKEN t)
{
	while(Token.type == '+' || Token.type == '-')
	{
		u16 type = Token.type;

		lexar();
		ST_TOKEN t2 = term();

		if( IsMatch(t2, INT) == TRUE )
		{
			t = Run(type, t, t2);
		}
	}

	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN term()
{
	ST_TOKEN t = factor();

	for(;;)
	{
		switch(Token.type)
		{
		case '*':
		case '/':
			return term2(t);
		default:
			return t;
		}
	}
}
//---------------------------------------------------------------------------
ST_TOKEN term2(ST_TOKEN t)
{
	while(Token.type == '*' || Token.type == '/')
	{
		u16 type = Token.type;
		lexar();

		if( IsMatch(Token, INT) == TRUE )
		{
			t = Run(type, t, Token);
		}
	}

	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN factor()
{
	ST_TOKEN t, t2;

	switch(Token.type)
	{
	case EOS: return Token;
	case ';': return Token;
	case INT:
		t = Token;
		lexar();
		return t;

	case '-':
		lexar();
		if( IsMatch(Token, INT) == TRUE )
		{
			Token.u.val *= -1;
			lexar();
			return Token;
		}
		Token.type = EOS;
		return Token;

	case '(':
		lexar();
		t = expr();
		if( IsMatch(Token, ')') == TRUE )
		{
			lexar();
			return t;
		}
		Token.type = EOS;
		return Token;

	case SYMBOL:
		t2 = Token;

		lexar();
		if( IsMatch(Token, '=') == TRUE )
		{
			lexar();
			t = expr();
			if( IsMatch(t, INT) == TRUE )
			{
				ChgSymbol(t2.u.str, t.u.val);
				return t;
			}
			t.type = EOS;
			return t;
		}
		t2.type  = INT;
		t2.u.val = GetSymbolVal(t2.u.str);
		return t2;

	case '!':
		lexar();
		t = expr();
		if( IsMatch(Token, INT) == TRUE )
		{
			Token.u.val = Token.u.val == 0 ? TRUE : FALSE;
			lexar();
			return Token;
		}
		Token.type = EOS;
		return Token;

	default:
		Token.type = EOS;
		return Token;
	}
}
//---------------------------------------------------------------------------
ST_TOKEN Run(u16 op, ST_TOKEN code1, ST_TOKEN code2)
{
	ST_TOKEN ret = code1;		//߂l̃x[Xcode1ɂ܂
	s16 val1 = code1.u.val;
	s16 val2 = code2.u.val;

	switch(op)
	{
	case '<': ret.u.val = val1 <  val2 ? TRUE : FALSE;  break;
	case '>': ret.u.val = val1 >  val2 ? TRUE : FALSE;  break;
	case EQ:  ret.u.val = val1 == val2 ? TRUE : FALSE;  break;
	case NE:  ret.u.val = val1 != val2 ? TRUE : FALSE;  break;
	case LE:  ret.u.val = val1 <= val2 ? TRUE : FALSE;  break;
	case GE:  ret.u.val = val1 >= val2 ? TRUE : FALSE;  break;

	case '+': ret.u.val = val1 + val2;  break;
	case '-': ret.u.val = val1 - val2;  break;
	case '*': ret.u.val = val1 * val2;  break;
	case '/': ret.u.val = val1 / val2;  break;

	default: break;
	}

	return ret;
}
//---------------------------------------------------------------------------
//s
void ProcBGLineFeed()
{
	ST_WRITER* wp = &Writer;

	//x
	if(wp->x + 1 >= BG_SCREEN_CX)
	{
		wp->x  = 0;
		wp->y += 1;
	}else{
		wp->x += 1;
	}

	//y
	if(wp->y >= BG_WRITE_EY)
	{
		wp->y = BG_WRITE_EY - 1;
		ProcScroll();
	}

	DrawWriter();
}
//---------------------------------------------------------------------------
void DrawAnswer(s32 ans)
{
	char msg[40];
	bprintf(msg, ":%d", ans);
	BGDrawStr(BG3.Map, 0, (Writer.y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, (u8*)msg);

	DrawWriter();
}
//---------------------------------------------------------------------------
bool WaitTimer()
{
	if(Timer.clockTick - Timer.wait < TIMER_INTMS)
	{
		return TRUE;
	}

	Timer.wait = Timer.clockTick;
	return FALSE;
}
//---------------------------------------------------------------------------
void InitToken()
{
	s16 i;

	//V{e[u
	for(i=0; i<MAX_SYMBOL_TABLE_CNT; i++)
	{
		SymTbl[i].type   = SYMBOL;
		SymTbl[i].str[0] = '\0';
		SymTbl[i].val    = 0;
	}
	SymRegistCnt = 0;

	//ߌe[u
	CmdTbl[0].type   = _IF;
	strcpy((char*)CmdTbl[0].str, (char*)"_if");

	CmdTbl[1].type   = _WHILE;
	strcpy((char*)CmdTbl[1].str, (char*)"_while");
}
//---------------------------------------------------------------------------
bool HasSymbol(u8* str)
{
	s16 i;

	for(i=0; i<SymRegistCnt; i++)
	{
		if( strcmp((char*)str, (char*)SymTbl[i].str) == 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
s32 GetSymbolVal(u8* str)
{
	s16 i;

	for(i=0; i<SymRegistCnt; i++)
	{
		if( strcmp((char*)str, (char*)SymTbl[i].str) == 0)
		{
			return SymTbl[i].val;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
u16 GetSymbolNum(u8* str)
{
	s16 i;

	for(i=0; i<SymRegistCnt; i++)
	{
		if( strcmp((char*)str, (char*)SymTbl[i].str) == 0)
		{
			return i;
		}
	}

	return 255;
}
//---------------------------------------------------------------------------
//HasSymbolsĂ
u8* GetSymbolPointer(u8* str)
{
	s16 i;

	for(i=0; i<SymRegistCnt; i++)
	{
		if( strcmp((char*)str, (char*)SymTbl[i].str) == 0 )
		{
			return SymTbl[i].str;
		}
	}

	return NULL;
}
//---------------------------------------------------------------------------
u8* AddSymbol(u8* str)
{
	u16 i = SymRegistCnt;
	
	SymTbl[i].type = SYMBOL;
	strcpy((char*)SymTbl[i].str, (char*)str);

	SymRegistCnt++;

	return SymTbl[i].str;
}
//---------------------------------------------------------------------------
//߂l ύX̗L
bool ChgSymbol(u8* str, s32 val)
{
	u16 i = GetSymbolNum(str);

	//o^̗L`FbNĂȂł

	SymTbl[i].val = val;
	return TRUE;
}
